# Copyright (C) 2020-2022 Benjamin Drung <bdrung@posteo.de>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Test Mmdebstrap class of bdebstrap."""

import logging
import os
import unittest
import unittest.mock
from unittest.mock import MagicMock

from bdebstrap import Config, Mmdebstrap, __script_name__


class TestMmdebstrap(unittest.TestCase):
    """
    This unittest class tests the Mmdebstrap object.
    """

    maxDiff = None

    def setUp(self) -> None:
        logging.getLogger(__script_name__).setLevel(logging.WARNING)

    def test_debian_example(self) -> None:
        """Test Mmdebstrap with Debian unstable config."""
        mmdebstrap = Mmdebstrap(
            Config(
                mmdebstrap={
                    "architectures": ["i386"],
                    "install-recommends": True,
                    "keyrings": ["/usr/share/keyrings/debian-archive-keyring.gpg"],
                    "mode": "unshare",
                    "suite": "unstable",
                    "target": "example.tar.xz",
                    "variant": "minbase",
                }
            )
        )
        self.assertEqual(
            mmdebstrap.construct_parameters("/output"),
            [
                "mmdebstrap",
                "--variant=minbase",
                "--mode=unshare",
                "--keyring=/usr/share/keyrings/debian-archive-keyring.gpg",
                "--architectures=i386",
                '--essential-hook=mkdir -p "$1/tmp/bdebstrap-output"',
                '--aptopt=Apt::Install-Recommends "true"',
                "--customize-hook=chroot \"$1\" dpkg-query -f='${Package}\\t${Version}\\n' -W "
                '> "$1/tmp/bdebstrap-output/manifest"',
                '--customize-hook=sync-out "/tmp/bdebstrap-output" "/output"',
                '--customize-hook=rm -rf "$1/tmp/bdebstrap-output"',
                "unstable",
                "example.tar.xz",
            ],
        )

    def test_dry_run(self) -> None:
        """Test Mmdebstrap with dry run set."""
        mmdebstrap = Mmdebstrap(
            Config(mmdebstrap={"suite": "unstable", "target": "example.tar.xz"})
        )
        self.assertEqual(
            mmdebstrap.construct_parameters("/output", True),
            [
                "mmdebstrap",
                "--simulate",
                '--essential-hook=mkdir -p "$1/tmp/bdebstrap-output"',
                "--customize-hook=chroot \"$1\" dpkg-query -f='${Package}\\t${Version}\\n' -W "
                '> "$1/tmp/bdebstrap-output/manifest"',
                '--customize-hook=sync-out "/tmp/bdebstrap-output" "/output"',
                '--customize-hook=rm -rf "$1/tmp/bdebstrap-output"',
                "unstable",
                "example.tar.xz",
            ],
        )

    def test_hooks(self) -> None:
        """Test Mmdebstrap with custom hooks."""
        mmdebstrap = Mmdebstrap(
            Config(
                mmdebstrap={
                    "cleanup-hooks": ['rm -f "$0/etc/udev/rules.d/70-persistent-net.rules"'],
                    "customize-hooks": [
                        'chroot "$0" update-alternatives --set editor /usr/bin/vim.basic'
                    ],
                    "essential-hooks": ["copy-in /etc/bash.bashrc /etc"],
                    "extract-hooks": ['find "$1" -xtype l'],
                    "hostname": "example",
                    "setup-hooks": [],
                    "suite": "buster",
                    "target": "buster.tar.xz",
                }
            )
        )
        self.assertEqual(
            mmdebstrap.construct_parameters("/output"),
            [
                "mmdebstrap",
                '--extract-hook=find "$1" -xtype l',
                '--essential-hook=mkdir -p "$1/tmp/bdebstrap-output"',
                "--essential-hook=copy-in /etc/bash.bashrc /etc",
                '--customize-hook=chroot "$0" update-alternatives --set editor /usr/bin/vim.basic',
                '--customize-hook=rm -f "$0/etc/udev/rules.d/70-persistent-net.rules"',
                '--customize-hook=echo "example" > "$1/etc/hostname"',
                "--customize-hook=chroot \"$1\" dpkg-query -f='${Package}\\t${Version}\\n' -W "
                '> "$1/tmp/bdebstrap-output/manifest"',
                '--customize-hook=sync-out "/tmp/bdebstrap-output" "/output"',
                '--customize-hook=rm -rf "$1/tmp/bdebstrap-output"',
                "buster",
                "buster.tar.xz",
            ],
        )

    def test_extra_opts(self) -> None:
        """Test Mmdebstrap with extra options."""
        mmdebstrap = Mmdebstrap(
            Config(
                mmdebstrap={
                    "aptopts": ['Acquire::http { Proxy "http://proxy:3128/"; }'],
                    "components": ["main", "non-free", "contrib"],
                    "dpkgopts": ["force-confdef", "force-confold"],
                    "format": "tar",
                    "hook-dirs": ["/usr/share/mmdebstrap/hooks/busybox"],
                    "packages": ["bash-completions", "vim"],
                    "suite": "unstable",
                    "skip": ["cleanup/apt", "update"],
                    "target": "example.tar.xz",
                }
            )
        )
        self.assertEqual(
            mmdebstrap.construct_parameters("/output"),
            [
                "mmdebstrap",
                "--format=tar",
                '--aptopt=Acquire::http { Proxy "http://proxy:3128/"; }',
                "--dpkgopt=force-confdef",
                "--dpkgopt=force-confold",
                "--include=bash-completions,vim",
                "--components=main,non-free,contrib",
                "--skip=cleanup/apt,update",
                "--hook-dir=/usr/share/mmdebstrap/hooks/busybox",
                '--essential-hook=mkdir -p "$1/tmp/bdebstrap-output"',
                "--customize-hook=chroot \"$1\" dpkg-query -f='${Package}\\t${Version}\\n' -W "
                '> "$1/tmp/bdebstrap-output/manifest"',
                '--customize-hook=sync-out "/tmp/bdebstrap-output" "/output"',
                '--customize-hook=rm -rf "$1/tmp/bdebstrap-output"',
                "unstable",
                "example.tar.xz",
            ],
        )

    @unittest.mock.patch("os.path.exists", unittest.mock.MagicMock(return_value=True))
    @unittest.mock.patch("os.stat")
    @unittest.mock.patch("os.utime")
    def test_clamp_mtime(self, utime_mock: MagicMock, stat_mock: MagicMock) -> None:
        """Test clamping mtime of output files/directories."""
        stat_mock.return_value = os.stat_result(
            (33261, 16535979, 64769, 1, 1000, 1000, 17081, 1581451059, 1581451059, 1581451059)
        )
        config = Config(
            env={"SOURCE_DATE_EPOCH": 1581433737}, mmdebstrap={"target": "/output/test.tar"}
        )
        mmdebstrap = Mmdebstrap(config)
        mmdebstrap.clamp_mtime("/output")
        self.assertEqual(utime_mock.call_count, 3)

    @unittest.mock.patch("os.path.exists", unittest.mock.MagicMock(return_value=True))
    @unittest.mock.patch("os.stat")
    @unittest.mock.patch("os.utime")
    def test_clamp_mtime_permission(self, utime_mock: MagicMock, stat_mock: MagicMock) -> None:
        """Test permission error when clamping mtime of output files/directories."""
        stat_mock.return_value = os.stat_result(
            (33261, 16535979, 64769, 1, 1000, 1000, 17081, 1581451059, 1581451059, 1581451059)
        )
        utime_mock.side_effect = PermissionError(1, "Operation not permitted")
        config = Config(
            env={"SOURCE_DATE_EPOCH": 1581433737}, mmdebstrap={"target": "/output/test.tar"}
        )
        mmdebstrap = Mmdebstrap(config)
        with self.assertLogs("bdebstrap", level="ERROR") as context_manager:
            mmdebstrap.clamp_mtime("/output")
            self.assertEqual(utime_mock.call_count, 3)
            self.assertEqual(
                [
                    "ERROR:bdebstrap:Failed to change modification time of '/output/manifest': "
                    "[Errno 1] Operation not permitted",
                    "ERROR:bdebstrap:Failed to change modification time of '/output/test.tar': "
                    "[Errno 1] Operation not permitted",
                    "ERROR:bdebstrap:Failed to change modification time of '/output': "
                    "[Errno 1] Operation not permitted",
                ],
                context_manager.output,
            )

    def test_log_level_debug(self) -> None:
        """Test Mmdebstrap with log level debug."""
        logging.getLogger(__script_name__).setLevel(logging.DEBUG)
        mmdebstrap = Mmdebstrap(Config())
        self.assertEqual(
            mmdebstrap.construct_parameters("/output")[0:2], ["mmdebstrap", "--debug"]
        )

    def test_log_level_error(self) -> None:
        """Test Mmdebstrap with log level error."""
        logging.getLogger(__script_name__).setLevel(logging.ERROR)
        mmdebstrap = Mmdebstrap(Config())
        self.assertEqual(mmdebstrap.construct_parameters("/output")[0:2], ["mmdebstrap", "-q"])

    def test_log_level_info(self) -> None:
        """Test Mmdebstrap with log level info."""
        logging.getLogger(__script_name__).setLevel(logging.INFO)
        mmdebstrap = Mmdebstrap(Config())
        self.assertEqual(mmdebstrap.construct_parameters("/output")[0:2], ["mmdebstrap", "-v"])

    def test_log_level_warning(self) -> None:
        """Test Mmdebstrap with log level warning."""
        logging.getLogger(__script_name__).setLevel(logging.WARNING)
        mmdebstrap = Mmdebstrap(Config())
        self.assertEqual(
            mmdebstrap.construct_parameters("/output")[0:2],
            ["mmdebstrap", '--essential-hook=mkdir -p "$1/tmp/bdebstrap-output"'],
        )



# Copyright (C) 2017-2021, Benjamin Drung <benjamin.drung@ionos.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Helper functions for testing."""

import inspect
import os
import unittest


def get_path(code_file: str) -> str:
    """Return relative or absolute path to given code file.

    During Debian package build, the current directory is
    changed to .pybuild/cpython3_*/build and the root directory
    is stored in the OLDPWD environment variable.
    """
    if os.path.exists(code_file):
        return code_file
    # The alternative path is needed for Debian's pybuild
    alternative = os.path.join(os.environ.get("OLDPWD", ""), code_file)
    if os.path.exists(alternative):
        return alternative
    return code_file


def get_source_files() -> list[str]:
    """Return a list of sources files/directories (to check with flake8/pylint)."""
    scripts = ["bdebstrap"]
    modules = ["tests"]
    py_files = ["setup.py"]

    files = []
    for code_file in scripts + modules + py_files:
        code_path = get_path(code_file)
        if code_file in scripts:
            with open(code_path, "rb") as script_file:
                shebang = script_file.readline().decode("utf-8")
            if "python" in shebang:
                files.append(code_path)
        else:
            files.append(code_path)
    return files


def unittest_verbosity() -> int:
    """
    Return the verbosity setting of the currently running unittest.

    If no test is running, return 0.
    """
    frame = inspect.currentframe()
    while frame:
        self = frame.f_locals.get("self")
        if isinstance(self, unittest.TestProgram):
            return self.verbosity
        frame = frame.f_back
    return 0  # pragma: no cover


class TestGetPath(unittest.TestCase):
    """Unittests for get_path function."""

    def test_get_path_exists(self) -> None:
        """Test get_path(__file__)."""
        path = get_path(__file__)
        self.assertEqual(path, __file__)

    def test_get_path_missing(self) -> None:
        """Test non-existing file for get_path()."""
        path = get_path(__file__ + "non-existing")
        self.assertEqual(path, __file__ + "non-existing")

    def test_get_path_pybuild(self) -> None:
        """Test changing current directory before calling get_path()."""
        relpath = os.path.relpath(__file__)
        oldpwd = os.environ.get("OLDPWD")
        curdir = os.getcwd()
        try:
            os.environ["OLDPWD"] = curdir
            os.chdir("/")
            path = get_path(relpath)
        finally:
            os.chdir(curdir)
            if oldpwd is not None:
                os.environ["OLDPWD"] = oldpwd
        self.assertEqual(os.path.normpath(path), __file__)



# Copyright (C) 2020-2022 Benjamin Drung <bdrung@posteo.de>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Test helper functions of bdebstrap."""

import os
import tempfile
import unittest
import unittest.mock
from unittest.mock import MagicMock

from bdebstrap import clamp_mtime, duration_str, escape_cmd, prepare_output_dir


class TestClampMtime(unittest.TestCase):
    """
    This unittest class tests the clamp_mtime function.
    """

    @staticmethod
    @unittest.mock.patch("os.stat")
    @unittest.mock.patch("os.utime")
    def test_clamp(utime_mock: MagicMock, stat_mock: MagicMock) -> None:
        """Test clamping the modification time."""
        stat_mock.return_value = os.stat_result(
            (33261, 16535979, 64769, 1, 1000, 1000, 17081, 1581451059, 1581451059, 1581451059)
        )
        clamp_mtime("/example", 1581433737)
        utime_mock.assert_called_once_with("/example", (1581433737, 1581433737))

    @staticmethod
    @unittest.mock.patch("os.stat")
    @unittest.mock.patch("os.utime")
    def test_not_clamping(utime_mock: MagicMock, stat_mock: MagicMock) -> None:
        """Test not clamping the modification time."""
        stat_mock.return_value = os.stat_result(
            (33261, 16535979, 64769, 1, 1000, 1000, 17081, 1581451059, 1581451059, 1581451059)
        )
        clamp_mtime("/example", 1581506399)
        utime_mock.assert_not_called()

    @staticmethod
    @unittest.mock.patch("os.stat")
    def test_no_source_date_epoch(stat_mock: MagicMock) -> None:
        """Test doing nothing if SOURCE_DATE_EPOCH is not set."""
        clamp_mtime("/example", None)
        stat_mock.assert_not_called()


class TestDuration(unittest.TestCase):
    """
    This unittest class tests the duration_str function.
    """

    def test_seconds(self) -> None:
        """Test calling duration_str(3.606104612350464)."""
        self.assertEqual(duration_str(3.606104612350464), "3.606 seconds")

    def test_minutes(self) -> None:
        """Test calling duration_str(421.88086652755737)."""
        self.assertEqual(duration_str(421.88086652755737), "7 min 1.881 s (= 421.881 s)")

    def test_hours(self) -> None:
        """Test calling duration_str(7397.447488069534)."""
        self.assertEqual(duration_str(7397.447488069534), "2 h 3 min 17.447 s (= 7397.447 s)")


class TestEscapeCmd(unittest.TestCase):
    """
    This unittest class tests the escape_cmd function.
    """

    def test_simple(self) -> None:
        """Test calling escape_cmd(["free"])."""
        self.assertEqual(escape_cmd(["free"]), "free")

    def test_spaces(self) -> None:
        """Test calling escape_cmd(["scp", "source", "a space"])."""
        self.assertEqual(escape_cmd(["scp", "source", "a space"]), 'scp source "a space"')

    def test_escape(self) -> None:
        """Test calling escape_cmd(["dpkg-query", r"-f=${Package}\t${Version}\n", "-W"])."""
        self.assertEqual(
            escape_cmd(["dpkg-query", r"-f=${Package}\t${Version}\n", "-W"]),
            r'dpkg-query "-f=\${Package}\t\${Version}\n" -W',
        )

    def test_customize_hook(self) -> None:
        """Test calling escape_cmd on mmdebstrap customize hook."""
        self.assertEqual(
            escape_cmd(
                [
                    "mmdebstrap",
                    '--customize-hook=chroot "$1" dpkg-query '
                    "-f='${Package}\\t${Version}\\n' -W > \"$1/tmp/bdebstrap/manifest\"",
                ]
            ),
            r'mmdebstrap "--customize-hook=chroot \"\$1\" dpkg-query '
            r"-f='${Package}\t${Version}\n' -W "
            r'> \"\$1/tmp/bdebstrap/manifest\""',
        )


class TestPrepareOutputDir(unittest.TestCase):
    """
    This unittest class tests the prepare_output_dir function.
    """

    TMP_PREFIX = "bdebstrap-"

    def test_dry_run(self) -> None:
        """Test that dry run does not create the directory."""
        with tempfile.TemporaryDirectory(prefix=self.TMP_PREFIX) as tmpdir:
            output_dir = os.path.join(tmpdir, "unstable")
            self.assertTrue(prepare_output_dir(output_dir, False, simulate=True))
            self.assertFalse(os.path.isdir(output_dir))

    def test_force(self) -> None:
        """Test replacing an existing output directory."""
        with tempfile.TemporaryDirectory(prefix=self.TMP_PREFIX) as tmpdir:
            output_dir = os.path.join(tmpdir, "unstable")
            os.makedirs(output_dir)
            os.mknod(os.path.join(output_dir, "root.tar"))

            self.assertTrue(prepare_output_dir(output_dir, True))
            self.assertTrue(os.path.isdir(output_dir))
            self.assertEqual(os.listdir(output_dir), [])

    def test_existing(self) -> None:
        """Test failure when output directory already exists."""
        with tempfile.TemporaryDirectory(prefix=self.TMP_PREFIX) as tmpdir:
            output_dir = os.path.join(tmpdir, "unstable")
            os.makedirs(output_dir)
            os.mknod(os.path.join(output_dir, "root.tar"))

            with self.assertLogs("bdebstrap", level="ERROR") as context_manager:
                self.assertFalse(prepare_output_dir(output_dir, False))
            self.assertIn("already exists and is not empty", context_manager.output[-1])
            self.assertTrue(os.path.isdir(output_dir))
            self.assertEqual(os.listdir(output_dir), ["root.tar"])

    def test_missing_output_dir(self) -> None:
        """Test creating the missing output directory."""
        with tempfile.TemporaryDirectory(prefix=self.TMP_PREFIX) as tmpdir:
            output_dir = os.path.join(tmpdir, "unstable")
            self.assertTrue(prepare_output_dir(output_dir, False))
            self.assertTrue(os.path.isdir(output_dir))



# Copyright (C) 2019-2022 Benjamin Drung <bdrung@posteo.de>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Test configuration handling of bdebstrap."""

import contextlib
import io
import logging
import os
import tempfile
import typing
import unittest
import unittest.mock

from bdebstrap import HOOKS_DIR, Config, dict_merge, parse_args

EXAMPLE_CONFIG_DIR = os.path.join(os.path.dirname(__file__), "..", "examples")
TEST_CONFIG_DIR = os.path.join(os.path.dirname(__file__), "configs")


def get_subset(dict_: typing.Any, keys: set[str]) -> dict[str, typing.Any]:
    """Return a dictionary that only contains the items for the given keys."""
    return {key: value for key, value in dict_.items() if key in keys}


class TestArguments(unittest.TestCase):
    """
    This unittest class tests the argument parsing.
    """

    maxDiff = None

    def test_debug(self) -> None:
        """Test --debug argument parsing."""
        args = parse_args(["--debug"])
        self.assertEqual(args.log_level, logging.DEBUG)

    def test_empty_args(self) -> None:
        """Test setting arguments to empty strings."""
        args = parse_args(
            [
                "--aptopt=",
                "--architectures=",
                "--cleanup-hook=",
                "--components=",
                "--config=",
                "--customize-hook=",
                "--dpkgopt=",
                "--essential-hook=",
                "--extract-hook=",
                "--hook-dir=",
                "--keyring=",
                "--skip=",
                "--mirrors=",
                "--packages=",
                "--setup-hook=",
            ]
        )
        self.assertEqual(
            get_subset(
                args.__dict__,
                {
                    "aptopt",
                    "architectures",
                    "cleanup_hook",
                    "components",
                    "config",
                    "customize_hook",
                    "dpkgopt",
                    "essential_hook",
                    "extract_hook",
                    "hook_dir",
                    "keyring",
                    "mirrors",
                    "packages",
                    "setup_hook",
                    "skip",
                },
            ),
            {
                "aptopt": [],
                "architectures": [],
                "cleanup_hook": [],
                "components": [],
                "config": [],
                "customize_hook": [],
                "dpkgopt": [],
                "essential_hook": [],
                "extract_hook": [],
                "hook_dir": [],
                "keyring": [],
                "mirrors": [],
                "packages": [],
                "setup_hook": [],
                "skip": [],
            },
        )

    def test_no_args(self) -> None:
        """Test calling bdebstrap without arguments."""
        args = parse_args([])
        self.assertEqual(
            args.__dict__,
            {
                "aptopt": None,
                "architectures": None,
                "cleanup_hook": None,
                "components": None,
                "config": [],
                "customize_hook": None,
                "dpkgopt": None,
                "env": {},
                "essential_hook": None,
                "extract_hook": None,
                "force": False,
                "format": None,
                "hook_dir": None,
                "hostname": None,
                "install_recommends": False,
                "keyring": None,
                "log_level": logging.WARNING,
                "mirrors": [],
                "mode": None,
                "name": None,
                "output_base_dir": ".",
                "output": None,
                "packages": None,
                "setup_hook": None,
                "simulate": False,
                "skip": None,
                "suite": None,
                "target": None,
                "tmpdir": None,
                "variant": None,
            },
        )

    def test_parse_env(self) -> None:
        """Test parsing --env parameters."""
        args = parse_args(["-e", "KEY=VALUE", "--env", "FOO=bar"])
        self.assertEqual(args.env, {"FOO": "bar", "KEY": "VALUE"})

    def test_malformed_env(self) -> None:
        """Test malformed --env parameter (missing equal sign)."""
        stderr = io.StringIO()
        with contextlib.redirect_stderr(stderr), self.assertRaises(SystemExit):
            parse_args(["--env", "invalid"])
        self.assertIn("Failed to parse --env 'invalid'.", stderr.getvalue())

    def test_mirrors_with_spaces(self) -> None:
        """Test --mirrors with leading/trailing spaces."""
        args = parse_args(
            [
                "--mirrors",
                "  deb http://deb.debian.org/debian unstable main\t ,  \t, "
                "deb http://deb.debian.org/debian unstable non-free\t",
                "--mirrors",
                "\tdeb http://deb.debian.org/debian unstable contrib ",
            ]
        )
        self.assertEqual(
            args.mirrors,
            [
                "deb http://deb.debian.org/debian unstable main",
                "deb http://deb.debian.org/debian unstable non-free",
                "deb http://deb.debian.org/debian unstable contrib",
            ],
        )

    def test_optional_args(self) -> None:
        """Test optional arguments (which also have positional ones)."""
        args = parse_args(
            [
                "--suite",
                "unstable",
                "--target",
                "unstable.tar",
                "--mirrors",
                "deb http://deb.debian.org/debian unstable main,"
                "deb http://deb.debian.org/debian unstable non-free",
                "--mirrors",
                "deb http://deb.debian.org/debian unstable contrib",
            ]
        )
        self.assertEqual(
            get_subset(args.__dict__, {"mirrors", "suite", "target"}),
            {
                "mirrors": [
                    "deb http://deb.debian.org/debian unstable main",
                    "deb http://deb.debian.org/debian unstable non-free",
                    "deb http://deb.debian.org/debian unstable contrib",
                ],
                "suite": "unstable",
                "target": "unstable.tar",
            },
        )

    def test_positional_args(self) -> None:
        """Test positional arguments (overwriting optional ones)."""
        args = parse_args(
            [
                "--suite",
                "bullseye",
                "--target",
                "bullseye.tar",
                "--mirrors",
                "deb http://deb.debian.org/debian unstable main,"
                "deb http://deb.debian.org/debian unstable non-free",
                "unstable",
                "unstable.tar",
                "deb http://deb.debian.org/debian unstable contrib",
            ]
        )
        self.assertEqual(
            get_subset(args.__dict__, {"mirrors", "suite", "target"}),
            {
                "mirrors": [
                    "deb http://deb.debian.org/debian unstable main",
                    "deb http://deb.debian.org/debian unstable non-free",
                    "deb http://deb.debian.org/debian unstable contrib",
                ],
                "suite": "unstable",
                "target": "unstable.tar",
            },
        )

    def test_split(self) -> None:
        """Test splitting comma and space separated values."""
        args = parse_args(
            [
                "--packages",
                "distro-info ionit,netconsole",
                "--include",
                "openssh-server,restricted-ssh-commands",
                "--components",
                "main,non-free contrib",
                "--architectures",
                "amd64,i386",
            ]
        )
        self.assertEqual(
            get_subset(args.__dict__, {"architectures", "components", "packages"}),
            {
                "architectures": ["amd64", "i386"],
                "components": ["main", "non-free", "contrib"],
                "packages": [
                    "distro-info",
                    "ionit",
                    "netconsole",
                    "openssh-server",
                    "restricted-ssh-commands",
                ],
            },
        )


class TestConfig(unittest.TestCase):
    """
    This unittest class tests the Config object.
    """

    maxDiff = None

    def test_add_command_line_arguments(self) -> None:
        """Test Config.add_command_line_arguments()."""
        args = parse_args(
            [
                "-c",
                os.path.join(EXAMPLE_CONFIG_DIR, "Debian-unstable.yaml"),
                "--name",
                "Debian-unstable",
            ]
        )
        config = Config()
        config.add_command_line_arguments(args)
        self.assertEqual(
            config,
            {
                "mmdebstrap": {
                    "keyrings": ["/usr/share/keyrings/debian-archive-keyring.gpg"],
                    "mode": "unshare",
                    "suite": "unstable",
                    "target": "root.tar.xz",
                    "variant": "minbase",
                },
                "name": "Debian-unstable",
            },
        )

    def test_config_and_arguments(self) -> None:
        """Test Config.add_command_line_arguments() with config file and arguments."""
        args = parse_args(
            [
                "-c",
                os.path.join(EXAMPLE_CONFIG_DIR, "Debian-unstable.yaml"),
                "--name",
                "Debian-unstable",
                "--variant",
                "standard",
                "--mode",
                "root",
                "--format",
                "tar",
                "--aptopt",
                'Apt::Install-Recommends "0"',
                "--keyring",
                "/usr/share/keyrings",
                "--dpkgopt",
                "force-confdef",
                "--dpkgopt",
                "force-confold",
                "--include",
                "ionit,netconsole",
                "--components",
                "main,non-free",
                "--architectures",
                "i386",
                "--mirrors",
                "http://deb.debian.org/debian",
                "unstable",
                "unstable.tar",
            ]
        )
        config = Config()
        config.add_command_line_arguments(args)
        self.assertDictEqual(
            config,
            {
                "mmdebstrap": {
                    "aptopts": ['Apt::Install-Recommends "0"'],
                    "architectures": ["i386"],
                    "components": ["main", "non-free"],
                    "dpkgopts": ["force-confdef", "force-confold"],
                    "format": "tar",
                    "keyrings": [
                        "/usr/share/keyrings/debian-archive-keyring.gpg",
                        "/usr/share/keyrings",
                    ],
                    "mirrors": ["http://deb.debian.org/debian"],
                    "mode": "root",
                    "packages": ["ionit", "netconsole"],
                    "suite": "unstable",
                    "target": "unstable.tar",
                    "variant": "standard",
                },
                "name": "Debian-unstable",
            },
        )

    def test_add_command_line_arguments_no_config(self) -> None:
        """Test Config.add_command_line_arguments() with no config file."""
        args = parse_args(
            [
                "--cleanup-hook",
                'cp /dev/null "$1/etc/hostname"',
                "--customize-hook",
                'chroot "$1" apt-get update',
                "--env",
                "KEY=VALUE",
                "--essential-hook",
                "copy-in /etc/bash.bashrc /etc",
                "--extract-hook",
                'find "$1" -xtype l',
                "--hostname",
                "cobb",
                "--install-recommends",
                "--skip=check/signed-by",
                "--hook-dir=/usr/share/mmdebstrap/hooks/eatmydata",
                "--name",
                "ubuntu-24.04",
                "--setup-hook",
                'echo root:x:0:0:root:/root:/bin/sh > "$1/etc/passwd"',
            ]
        )
        config = Config()
        config.add_command_line_arguments(args)
        self.assertEqual(
            config,
            {
                "env": {"KEY": "VALUE"},
                "mmdebstrap": {
                    "cleanup-hooks": ['cp /dev/null "$1/etc/hostname"'],
                    "customize-hooks": ['chroot "$1" apt-get update'],
                    "essential-hooks": ["copy-in /etc/bash.bashrc /etc"],
                    "extract-hooks": ['find "$1" -xtype l'],
                    "hook-dirs": ["/usr/share/mmdebstrap/hooks/eatmydata"],
                    "hostname": "cobb",
                    "install-recommends": True,
                    "setup-hooks": ['echo root:x:0:0:root:/root:/bin/sh > "$1/etc/passwd"'],
                    "skip": ["check/signed-by"],
                },
                "name": "ubuntu-24.04",
            },
        )

    @staticmethod
    def test_check_example() -> None:
        """Test example unstable.yaml file."""
        config = Config()
        config.load(os.path.join(EXAMPLE_CONFIG_DIR, "Debian-unstable.yaml"))
        config["name"] = "Debian-unstable"
        config.check()

    @staticmethod
    def test_commented_packages() -> None:
        """Test commented-packages.yaml file."""
        config = Config()
        config.load(os.path.join(TEST_CONFIG_DIR, "commented-packages.yaml"))
        config.sanitize_packages()
        config.check()

    def test_env_items(self) -> None:
        """Test environment variables for example unstable.yaml."""
        config = Config()
        config.load(os.path.join(EXAMPLE_CONFIG_DIR, "Debian-unstable.yaml"))
        config["name"] = "Debian-unstable"
        self.assertEqual(
            config.env_items(),
            [
                ("BDEBSTRAP_HOOKS", HOOKS_DIR),
                ("BDEBSTRAP_NAME", "Debian-unstable"),
                ("BDEBSTRAP_OUTPUT_DIR", "/tmp/bdebstrap-output"),
            ],
        )

    def test_loading(self) -> None:
        """Test loading a YAML configuration file."""
        config = Config()
        config.load(os.path.join(EXAMPLE_CONFIG_DIR, "Debian-unstable.yaml"))
        self.assertEqual(
            config,
            {
                "mmdebstrap": {
                    "keyrings": ["/usr/share/keyrings/debian-archive-keyring.gpg"],
                    "mode": "unshare",
                    "suite": "unstable",
                    "target": "root.tar.xz",
                    "variant": "minbase",
                }
            },
        )

    def test_sanitize_packages_debs(self) -> None:
        """Test sanitize_packages method: multiple local .debs"""
        config = Config()
        config["mmdebstrap"] = {"packages": ["/home/user/foo.deb", "/home/user/bar.deb"]}
        config.sanitize_packages()
        self.assertEqual(
            config["mmdebstrap"]["packages"], ["/home/user/foo.deb", "/home/user/bar.deb"]
        )

    def test_sanitize_packages_duplicate_debs(self) -> None:
        """Test sanitize_packages method: remove duplicate local .debs."""
        config = Config()
        config["mmdebstrap"] = {
            "packages": ["./bdebstrap_0.5_all.deb", "../bdebstrap_0.4_all.deb"]
        }
        config.sanitize_packages()
        self.assertEqual(config["mmdebstrap"]["packages"], ["../bdebstrap_0.4_all.deb"])

    def test_sanitize_packages_duplicates(self) -> None:
        """Test sanitize_packages method: remove duplicates."""
        config = Config()
        config["mmdebstrap"] = {"packages": ["less/jammy-updates", "more", "less=590-1build1"]}
        config.sanitize_packages()
        self.assertEqual(config["mmdebstrap"]["packages"], ["less=590-1build1", "more"])

    def test_sanitize_packages_pattern(self) -> None:
        """Test sanitize_packages method: APT pattern"""
        config = Config()
        config["mmdebstrap"] = {"packages": ["?priority(required)", "?priority(important)"]}
        config.sanitize_packages()
        self.assertEqual(
            config["mmdebstrap"]["packages"], ["?priority(required)", "?priority(important)"]
        )

    def test_yaml_rendering(self) -> None:
        """Test that config.yaml is a syntactically valid yaml file."""
        config = Config()
        config_filename = os.path.join(EXAMPLE_CONFIG_DIR, "Debian-unstable.yaml")
        config.load(config_filename)
        with tempfile.NamedTemporaryFile() as temp_file:
            config.save(temp_file.name)
            with open(temp_file.name, encoding="utf-8") as config_file:
                output_config = config_file.read()
        with open(config_filename, encoding="utf-8") as config_file:
            input_config = config_file.read()
        self.assertEqual(output_config, input_config)

    def test_yaml_flow_style(self) -> None:
        """Test that config.yaml follows the correct flow style."""
        config = Config()
        config["mmdebstrap"] = {"packages": ["1", "2", "3", "4", "5"]}
        with tempfile.NamedTemporaryFile() as temp_file:
            config.save(temp_file.name)
            with open(temp_file.name, encoding="utf-8") as config_file:
                lines = len(config_file.readlines())
        self.assertEqual(lines, 8)

    def test_source_date_epoch(self) -> None:
        """Test getting and setting SOURCE_DATE_EPOCH."""
        config = Config()
        self.assertIsNone(config.source_date_epoch)
        with unittest.mock.patch("time.time", return_value=1581694618.0388665):
            config.set_source_date_epoch()
        self.assertEqual(config.source_date_epoch, 1581694618)

    def test_wrong_element_type(self) -> None:
        """Test error message for wrong list element type."""
        config = Config()
        config.load(os.path.join(TEST_CONFIG_DIR, "wrong-element-type.yaml"))
        with self.assertRaisesRegex(ValueError, "'customize-hooks' has type 'CommentedMap'"):
            config.check()


class TestDictMerge(unittest.TestCase):
    """Unittests for dict_merge function."""

    def test_merge_lists(self) -> None:
        """Test merging nested dicts."""
        items = {"A": ["A1", "A2", "A3"], "C": 4}
        dict_merge(items, {"A": ["A4", "A5"]})
        self.assertEqual(items, {"A": ["A1", "A2", "A3", "A4", "A5"], "C": 4})

    def test_merge_nested_dicts(self) -> None:
        """Test merging nested dicts."""
        items = {"A": {"A1": 0, "A4": 4}, "C": 4}
        dict_merge(items, {"A": {"A1": 1, "A5": 5}})
        self.assertEqual(items, {"A": {"A1": 1, "A4": 4, "A5": 5}, "C": 4})



# Copyright (C) 2021-2022, Benjamin Drung <bdrung@posteo.de>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Run shellcheck on Shell code."""

import os
import subprocess
import sys
import unittest

from . import get_path, unittest_verbosity

SHELL_SCRIPTS = ["hooks/disable-units", "hooks/enable-units"]


class ShellcheckTestCase(unittest.TestCase):
    """
    This unittest class provides a test that runs the shellcheck
    on Shell source code.
    """

    @unittest.skipIf(os.environ.get("SKIP_LINTERS"), "requested via SKIP_LINTERS env variable")
    def test_shellcheck(self) -> None:
        """Test: Run shellcheck on Shell source code."""
        cmd = ["shellcheck"] + [get_path(s) for s in SHELL_SCRIPTS]
        if unittest_verbosity() >= 2:
            sys.stderr.write(f"Running following command:\n{' '.join(cmd)}\n")
        with subprocess.Popen(
            cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True
        ) as process:
            out, err = process.communicate()

        if process.returncode != 0:  # pragma: no cover
            msgs = []
            if err:
                msgs.append(
                    f"shellcheck exited with code {process.returncode} "
                    f"and has unexpected output on stderr:\n{err.decode().rstrip()}"
                )
            if out:
                msgs.append(f"shellcheck found issues:\n{out.decode().rstrip()}")
            if not msgs:
                msgs.append(
                    f"shellcheck exited with code {process.returncode} "
                    "and has no output on stdout or stderr."
                )
            self.fail("\n".join(msgs))



# Copyright (C) 2010, Stefano Rivera <stefanor@debian.org>
# Copyright (C) 2017-2018, Benjamin Drung <bdrung@debian.org>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.

"""Run pylint."""

import os
import re
import subprocess
import sys
import unittest

from . import get_source_files, unittest_verbosity

CONFIG = os.path.join(os.path.dirname(__file__), "..", ".pylintrc")


class PylintTestCase(unittest.TestCase):
    """
    This unittest class provides a test that runs the pylint code check
    on the Python source code. The list of source files is provided by
    the get_source_files() function and pylint is purely configured via
    a config file.
    """

    def test_pylint(self) -> None:
        """Test: Run pylint on Python source code."""
        cmd = ["pylint", "--rcfile=" + CONFIG, "--"] + get_source_files()
        if os.environ.get("SKIP_LINTERS"):
            cmd.insert(2, "--errors-only")
        if unittest_verbosity() >= 2:
            sys.stderr.write(f"Running following command:\n{' '.join(cmd)}\n")
        process = subprocess.run(cmd, capture_output=True, check=False, text=True)

        if process.returncode != 0:  # pragma: no cover
            # Strip trailing summary (introduced in pylint 1.7). This summary might look like:
            #
            # ------------------------------------
            # Your code has been rated at 10.00/10
            #
            out = re.sub(
                "^(-+|Your code has been rated at .*)$", "", process.stdout, flags=re.MULTILINE
            ).rstrip()

            # Strip logging of used config file (introduced in pylint 1.8)
            err = re.sub("^Using config file .*\n", "", process.stderr.rstrip())

            msgs = []
            if err:
                msgs.append(
                    f"pylint exited with code {process.returncode} "
                    f"and has unexpected output on stderr:\n{err}"
                )
            if out:
                msgs.append(f"pylint found issues:\n{out}")
            if not msgs:
                msgs.append(
                    f"pylint exited with code {process.returncode} "
                    "and has no output on stdout or stderr."
                )
            self.fail("\n".join(msgs))



# Copyright (C) 2021-2022 Benjamin Drung <bdrung@posteo.de>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Test main function of bdebstrap."""

import os
import subprocess
import unittest
import unittest.mock

from bdebstrap import OUTPUT_DIR, main

EXAMPLE_CONFIG_DIR = os.path.join(os.path.dirname(__file__), "..", "examples")


def default_hooks(output_dir: str) -> list[str]:
    """Return the list of default hooks."""
    return [
        f'--essential-hook=mkdir -p "$1{OUTPUT_DIR}"',
        "--customize-hook=chroot \"$1\" dpkg-query -f='${Package}\\t${Version}\\n' -W "
        f'> "$1{OUTPUT_DIR}/manifest"',
        f'--customize-hook=sync-out "{OUTPUT_DIR}" "{output_dir}"',
        f'--customize-hook=rm -rf "$1{OUTPUT_DIR}"',
    ]


class TestMain(unittest.TestCase):
    """
    This unittest class tests the main function.
    """

    @unittest.mock.patch("bdebstrap.Config.save")
    @unittest.mock.patch("bdebstrap.prepare_output_dir")
    @unittest.mock.patch("bdebstrap.Mmdebstrap.call")
    def test_debian_example(
        self,
        mmdebstrap_call_mock: unittest.mock.MagicMock,
        prepare_output_dir_mock: unittest.mock.MagicMock,
        config_save_mock: unittest.mock.MagicMock,
    ) -> None:
        """Test Debian unstable example."""
        args = [
            "-c",
            os.path.join(EXAMPLE_CONFIG_DIR, "Debian-unstable.yaml"),
            "--name",
            "Debian-unstable",
        ]
        with self.assertLogs("bdebstrap", level="INFO") as context_manager:
            self.assertEqual(main(args), 0)
            self.assertIn("Execution time", context_manager.output[-1])
        config_save_mock.assert_called_once_with("./Debian-unstable/config.yaml", False)
        mmdebstrap_call_mock.assert_called_once_with("./Debian-unstable", False)
        prepare_output_dir_mock.assert_called_once_with("./Debian-unstable", False, False)

    @unittest.mock.patch("bdebstrap.Config.save")
    @unittest.mock.patch("bdebstrap.prepare_output_dir")
    @unittest.mock.patch("bdebstrap.Mmdebstrap.call")
    def test_failed_mmdebstrap(
        self,
        mmdebstrap_call_mock: unittest.mock.MagicMock,
        prepare_output_dir_mock: unittest.mock.MagicMock,
        config_save_mock: unittest.mock.MagicMock,
    ) -> None:
        """Test failure of mmdebstrap call."""
        mmdebstrap_call_mock.side_effect = subprocess.CalledProcessError(42, "mmdebstrap")
        args = ["-c", os.path.join(EXAMPLE_CONFIG_DIR, "Debian-unstable.yaml"), "--name", "foobar"]
        with self.assertLogs("bdebstrap", level="ERROR") as context_manager:
            self.assertEqual(main(args), 1)
            self.assertEqual(
                ["ERROR:bdebstrap:mmdebstrap failed with exit code 42. See above for details."],
                context_manager.output,
            )
        config_save_mock.assert_called_once_with("./foobar/config.yaml", False)
        mmdebstrap_call_mock.assert_called_once_with("./foobar", False)
        prepare_output_dir_mock.assert_called_once_with("./foobar", False, False)

    @unittest.mock.patch("bdebstrap.Config.save")
    @unittest.mock.patch("bdebstrap.prepare_output_dir")
    @unittest.mock.patch("bdebstrap.Mmdebstrap.call")
    def test_empty_target(
        self,
        mmdebstrap_call_mock: unittest.mock.MagicMock,
        prepare_output_dir_mock: unittest.mock.MagicMock,
        config_save_mock: unittest.mock.MagicMock,
    ) -> None:
        """Test keeping target empty."""
        with self.assertLogs("bdebstrap", level="INFO"):
            self.assertEqual(main(["--name", "empty-target", "unstable"]), 0)
        config_save_mock.assert_called_once_with("./empty-target/config.yaml", False)
        mmdebstrap_call_mock.assert_called_once_with("./empty-target", False)
        prepare_output_dir_mock.assert_called_once_with("./empty-target", False, False)

    @unittest.mock.patch("bdebstrap.Config.save")
    @unittest.mock.patch("bdebstrap.prepare_output_dir")
    @unittest.mock.patch("subprocess.check_call")
    def test_minus_target(
        self,
        check_call_mock: unittest.mock.MagicMock,
        prepare_output_dir_mock: unittest.mock.MagicMock,
        config_save_mock: unittest.mock.MagicMock,
    ) -> None:
        """Test --target=-."""
        with self.assertLogs("bdebstrap", level="INFO"):
            self.assertEqual(main(["--target=-", "--name", "minus-target", "unstable"]), 0)
        check_call_mock.assert_called_once_with(
            ["mmdebstrap", "-v"] + default_hooks("./minus-target") + ["unstable", "-"]
        )
        config_save_mock.assert_called_once_with("./minus-target/config.yaml", False)
        prepare_output_dir_mock.assert_called_once_with("./minus-target", False, False)



# Copyright (C) 2023, Benjamin Drung <bdrung@ubuntu.com>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
# OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.

"""Run mypy to check static typing of the Python code."""

import subprocess
import sys
import unittest

import ruamel.yaml

from . import get_source_files, unittest_verbosity


class MypyTestCase(unittest.TestCase):
    """
    This unittest class provides a test that runs mypy to check static
    typing of the Python code. The list of source files is provided by
    the get_source_files() function.
    """

    def test_mypy(self) -> None:
        """Test: Run mypy on Python source code."""
        if [int(x) for x in ruamel.yaml.__version__.split(".")] < [0, 18]:
            self.skipTest("false positives with ruamel.yaml < 0.18")
        cmd = ["mypy", "--strict"] + get_source_files()
        if unittest_verbosity() >= 2:
            sys.stderr.write(f"Running following command:\n{' '.join(cmd)}\n")
        process = subprocess.run(cmd, capture_output=True, check=False, text=True)

        if process.returncode != 0:  # pragma: no cover
            msgs = []
            if process.stderr:
                msgs.append(
                    f"mypy exited with code {process.returncode} and has"
                    f" unexpected output on stderr:\n{process.stderr.rstrip()}"
                )
            if process.stdout:
                msgs.append(f"mypy found issues:\n{process.stdout.rstrip()}")
            if not msgs:
                msgs.append(
                    f"mypy exited with code {process.returncode} "
                    "and has no output on stdout or stderr."
                )
            self.fail("\n".join(msgs))



#!/usr/bin/python3

# Copyright (C) 2019-2022, Benjamin Drung <bdrung@posteo.de>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Setup for bdebstrap"""

import logging
import os
import subprocess

from setuptools import Command, setup

try:
    from setuptools.command.build import build
except ImportError:
    # Fallback for setuptools < 60 and Python < 3.12
    # pylint: disable-next=deprecated-module
    from distutils.command.build import build  # type: ignore

HOOKS = ["hooks/disable-units", "hooks/enable-units"]
MAN_PAGES = ["bdebstrap.1"]


class DocCommand(Command):
    """A custom command to build the documentation using pandoc."""

    description = "run pandoc to generate man pages"
    user_options: list[tuple[str, str, str]] = []

    def initialize_options(self) -> None:
        """Set default values for options."""

    def finalize_options(self) -> None:
        """Post-process options."""

    # pylint: disable-next=no-self-use
    def run(self) -> None:
        """Run pandoc."""
        logger = logging.getLogger(__name__)
        for man_page in MAN_PAGES:
            command = ["pandoc", "-s", "-t", "man", man_page + ".md", "-o", man_page]
            logger.info("running command: %s", " ".join(command))
            subprocess.check_call(command)


# pylint: disable-next=too-few-public-methods
class BuildCommand(build):
    """Custom build command (calling doc beforehand)."""

    def run(self) -> None:
        """Execute the build command."""
        self.run_command("doc")
        super().run()


class CleanCommand(Command):
    """Custom clean command (removing generated man pages)."""

    description = "remove generated man pages"
    user_options = [("all", "a", "remove all build output, not just temporary by-products")]
    boolean_options = ["all"]

    def initialize_options(self) -> None:
        """Set default values for options."""

    def finalize_options(self) -> None:
        """Post-process options."""

    # pylint: disable-next=no-self-use
    def run(self) -> None:
        """Clean build artefacts from doc command."""
        logger = logging.getLogger(__name__)
        for man_page in MAN_PAGES:
            if os.path.exists(man_page):
                logger.info("removing %s", man_page)
                os.remove(man_page)


if __name__ == "__main__":
    with open("README.md", "r", encoding="utf-8") as fh:
        LONG_DESCRIPTION = fh.read()

    setup(
        name="bdebstrap",
        version="0.7.0",
        description="Benjamin's multi-mirror Debian chroot creation tool",
        long_description=LONG_DESCRIPTION,
        long_description_content_type="text/markdown",
        author="Benjamin Drung",
        author_email="bdrung@posteo.de",
        url="https://github.com/bdrung/bdebstrap",
        project_urls={"Bug Tracker": "https://github.com/bdrung/bdebstrap/issues"},
        license="ISC",
        classifiers=[
            "Development Status :: 5 - Production/Stable",
            "Environment :: Console",
            "License :: OSI Approved :: ISC License (ISCL)",
            "Operating System :: POSIX",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3 :: Only",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
        ],
        cmdclass={"doc": DocCommand, "build": BuildCommand, "clean": CleanCommand},
        install_requires=["ruamel.yaml"],
        scripts=["bdebstrap"],
        py_modules=[],
        data_files=[("/usr/share/man/man1", MAN_PAGES), ("/usr/share/bdebstrap/hooks", HOOKS)],
    )



# Copyright (C) 2017-2018, Benjamin Drung <bdrung@debian.org>
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Run flake8 check."""

import os
import subprocess
import sys
import unittest

from . import get_source_files, unittest_verbosity


class Flake8TestCase(unittest.TestCase):
    """
    This unittest class provides a test that runs the flake8 code
    checker (which combines pycodestyle and pyflakes) on the Python
    source code. The list of source files is provided by the
    get_source_files() function.
    """

    @unittest.skipIf(os.environ.get("SKIP_LINTERS"), "requested via SKIP_LINTERS env variable")
    def test_flake8(self) -> None:
        """Test: Run flake8 on Python source code."""
        cmd = [sys.executable, "-m", "flake8", "--max-line-length=99"] + get_source_files()
        if unittest_verbosity() >= 2:
            sys.stderr.write(f"Running following command:\n{' '.join(cmd)}\n")
        process = subprocess.run(cmd, capture_output=True, check=False, text=True)

        if process.returncode != 0:  # pragma: no cover
            msgs = []
            if process.stderr:
                msgs.append(
                    f"flake8 exited with code {process.returncode} and has"
                    f" unexpected output on stderr:\n{process.stderr.rstrip()}"
                )
            if process.stdout:
                msgs.append(f"flake8 found issues:\n{process.stdout.rstrip()}")
            if not msgs:
                msgs.append(
                    f"flake8 exited with code {process.returncode} "
                    "and has no output on stdout or stderr."
                )
            self.fail("\n".join(msgs))
